﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using parkMonitor.DB;
using parkMonitor.LOG;
using parkMonitor.entity;
using parkMonitor.model;
using System.Configuration;
using System.Threading;
using parkMonitor.server.uiLogServer;

namespace parkMonitor.server.CoreThread
{
    public abstract class AbstractCmd
    {
        public int equipmentStatus_address { set; get; }
        public int park_completed_address { set; get; }
        public int parkingEntX { set; get; }
        public int parkingEntY { set; get; }
        public int parkingEntZ { set; get; }
        public int fetch_completed_address { set; get; }
        public IEquipments PLC { set; get; }
        public IEquipments queuingThread { set; get; }
        public DBOperation oper { set; get; }
        public IEquipments NumMachine { set; get; }
        public CEntrance c { set; get; }
        public AllotParkingSpace aps { set; get; }
        public abstract void executeCmd(Command queueCmd);

        public AbstractCmd()
        {
            equipmentStatus_address = Convert.ToInt32(ConfigurationManager.AppSettings["equipmentStatus_address"]);
            park_completed_address = Convert.ToInt32(ConfigurationManager.AppSettings["park_completed_address"]);
            parkingEntX = Convert.ToInt32(ConfigurationManager.AppSettings["parkingEntX"]);
            parkingEntY = Convert.ToInt32(ConfigurationManager.AppSettings["parkingEntY"]);
            parkingEntZ = Convert.ToInt32(ConfigurationManager.AppSettings["parkingEntZ"]);
            //取车完成
            fetch_completed_address = Convert.ToInt32(ConfigurationManager.AppSettings["fetch_completed_address"]);
            //获取PLC句柄
            PLC = EquipmentSimpleFactory.ins.FindEquipment(EquipmentName.PLC);
            //获取队列句柄
            queuingThread = EquipmentSimpleFactory.ins.FindEquipment(EquipmentName.Queue);
            //获取号牌机线程句柄
            NumMachine = EquipmentSimpleFactory.ins.FindEquipment(EquipmentName.NumMachine);
            //数据库
            oper = new DBOperation();
            //车位分配
            c = new CEntrance();
            aps = new AllotParkingSpace();
        }

    }

    public class StopCmd : AbstractCmd
    {
        /// <summary>
        /// 命令中号牌与对应号牌机中当前号牌比对，确认车辆
        /// </summary>
        public bool NumValidation(Command queueCmd)
        {
            List<NumberMachineNode> numList = ((NumberMachineMessage)NumMachine.GetMessage()).data;
            bool result = false;
            Task numberCheck = Task.Factory.StartNew(() =>
            {
                int countdown = 8, count = 0;
                while (countdown-- > 0)
                {
                    foreach (NumberMachineNode node in numList)
                    {
                        if (queueCmd.ip.Equals(node.ip) && queueCmd.LicenseNum.Equals(node.LicenseNum))
                        {
                            count += 1;
                        }
                    }
                    Thread.Sleep(500);
                }
                if (count >= (int)(countdown * 0.8))
                {
                    result = true;
                }
                else
                {
                    result = false;
                }
            });
            numberCheck.Wait();
            return result;
        }

        public override void executeCmd(Command queueCmd)
        {
            int centerX = 0;
            int centerY = 0;
            int angleA = 0;
            int length = 0;
            int width = 0;
            int height = 0;
            PLCMessage PLCMsg = null;
            int status = 1;//停车

            //号牌验证,异常则退回队列
            Func<Command, bool> numValidationResult = new Func<Command, bool>(NumValidation);
            IAsyncResult result = numValidationResult.BeginInvoke(queueCmd, null, null);
            bool checkResult = numValidationResult.EndInvoke(result);
            if (!checkResult)
            {
                queuingThread.SetMessage(queueCmd);
                return;
            }

            ParkingMessage pm = new ParkingMessage();
            pm.status = status;
            PLC.SetMessage(pm);
            Console.WriteLine("停车，开激光:" + queueCmd.LicenseNum);
            Log.WriteLog("停车，开激光:" + queueCmd.LicenseNum);
            UILogServer.ins.info("停车，开激光:" + queueCmd.LicenseNum);
            c.parkingEntX = parkingEntX;
            c.parkingEntY = parkingEntY;
            c.parkingEntZ = parkingEntZ;
            Parking_Space ppp = new Parking_Space();
            Task parkingSpace = Task.Factory.StartNew(() =>
            {
                while (true)
                {
                    ppp = aps.MallocParkingSpace(c);//得到车位
                    if (ppp != null)
                    {
                        break;
                    }
                    Thread.Sleep(200);
                }
            });
            parkingSpace.Wait();
            int parkingSpaceID = ppp.parkingSpaceID;
            int parkingSpaceX = ppp.parkingSpaceX;
            int parkingSpaceY = ppp.parkingSpaceY;
            int parkingSpaceZ = ppp.parkingSpaceZ;
            //int garageID = ppp.garageID;
            string numberPlate = queueCmd.LicenseNum;
            int userID = 0;
            //过期号牌判断
            if (queueCmd.userID != "")
            {
                userID = Convert.ToInt32(queueCmd.userID);
            }
            int garageID = queueCmd.garageID;
            string realParkTime = queueCmd.TimeRecord;

            Task dataFromLaser = Task.Factory.StartNew(() =>
            {
                while (true)
                {
                    bool jumpOut = false;
                    PLCMsg = (PLCMessage)PLC.GetMessage();
                    if (PLCMsg != null)
                    {
                        foreach (LaserMessage lm in PLCMsg.laserMsgList)
                        {
                            if(lm.id == 1 && lm.data!=null && lm.status == 3)
                            {
                                centerX = lm.data.centerX;
                                centerY = lm.data.centerY;
                                angleA = lm.data.angleA;
                                length = lm.data.length;
                                width = lm.data.width;
                                height = lm.data.height;
                                jumpOut = true;
                                break;
                            }else if (lm.status == -1)
                            {
                                LogFile logFile = LogFile.ERROR;
                                Log.WriteLog(logFile, "激光异常");
                            }
                        }
                        if (jumpOut)
                        {
                            break;
                        }
                    }
                    Thread.Sleep(500);
                }
            });
            //Task.WaitAll(dataFromLaser,parkingSpace);
            dataFromLaser.Wait();

            //号牌验证,异常则退回队列
            numValidationResult = new Func<Command, bool>(NumValidation);
            result = numValidationResult.BeginInvoke(queueCmd, null, null);
            checkResult = numValidationResult.EndInvoke(result);
            if (!checkResult)
            {
                queuingThread.SetMessage(queueCmd);
                return;
            }

            status = 2;
            pm.status = status;
            pm.centerX = Convert.ToString(centerX);
            pm.centerY = Convert.ToString(centerY);
            pm.angleA = Convert.ToString(angleA);
            pm.length = Convert.ToString(length);
            pm.width = Convert.ToString(width);
            pm.height = Convert.ToString(height);
            pm.parkingSpaceID = Convert.ToString(parkingSpaceID);
            pm.parkingSpaceX = Convert.ToString(parkingSpaceX);
            pm.parkingSpaceY = Convert.ToString(parkingSpaceY);
            pm.parkingSpaceZ = Convert.ToString(parkingSpaceZ);
            PLC.SetMessage(pm);

            Console.WriteLine("停车，抓车:" + queueCmd.LicenseNum);
            Log.WriteLog("停车，抓车:" + queueCmd.LicenseNum);
            UILogServer.ins.info("停车，抓车:" + queueCmd.LicenseNum);
            Task storeSignal = Task.Factory.StartNew(() =>
            {
                while (true)
                {
                    Thread.Sleep(1000);
                    PLCMsg = (PLCMessage)PLC.GetMessage();
                    int storeStatus = Convert.ToInt32(PLCMsg.plcList[park_completed_address].Value);
                    //停车完成信号
                    if (storeStatus == 1)
                    {
                        break;
                    }
                }
            });
            storeSignal.Wait();
            status = 3;
            pm.status = status;
            PLC.SetMessage(pm);

            //停车完成，将相应车牌复位
            NumberMachineMessage numberMachineMessage = new NumberMachineMessage();
            numberMachineMessage.aNode = new NumberMachineNode();
            numberMachineMessage.aNode.ip = "";
            numberMachineMessage.aNode.LicenseNum = queueCmd.LicenseNum;
            queuingThread.SetMessage(numberMachineMessage);

            Console.WriteLine("停车完成，状态复位");
            Log.WriteLog("停车完成，状态复位");
            UILogServer.ins.info("停车完成，状态复位");
            //根据号牌查找车型
            //int vehicleTypeID = oper.getVehicleTypeID(numberPlate);
            //判断测量数据是否准确
            //bool isDataRight = oper.IsDataRight(vehicleTypeLength, vehicleTypeWidth, vehicleTypeHeight, vehicleTypeWheelbase, vehicleTypeID);

            //正常写入数据库
            if (userID != 0)
            {
                //更新数据库
                //插入停车记录表
                int parkingRecordsID = oper.InsertToParkingRecords(userID, numberPlate, parkingSpaceID, garageID, 3, realParkTime);
                //车辆表更新车辆信息
                oper.UpdateVehicle(numberPlate, 1, realParkTime, parkingRecordsID, parkingSpaceID, 1);
                //更新车库表剩余车位数
                int freeSpaceCount = oper.getGarageFreeSpace(garageID);
                freeSpaceCount = freeSpaceCount - 1;
                oper.UpdateGarageFreeSpace(freeSpaceCount, garageID);
                //更新车位表车位状态
                oper.UpdateParkingSpaceState(parkingSpaceID, 1);
                //插入消息队列表
                oper.InsertToMessageQueue(userID, "停车成功", 1);
            }
            //异常写入日志文件
            else
            {
                LogFile logFile = LogFile.ERROR_NUMBERPLATE;
                Log.WriteLog(logFile, "号牌：" + numberPlate);
                Log.WriteLog(logFile, "入库时间：" + realParkTime);
                Log.WriteLog(logFile, "车位id：" + parkingSpaceID);
                Log.WriteLog(logFile, "车位x：" + parkingSpaceX);
                Log.WriteLog(logFile, "车位y：" + parkingSpaceY);
                Log.WriteLog(logFile, "车位z：" + parkingSpaceZ);
                //异常情况处理有待讨论
            }
        }
    }

    public class FetchCmd : AbstractCmd
    {
        public override void executeCmd(Command queueCmd)
        {
            PLCMessage PLCMsg = null;
            Vehicle vehiclelist = oper.GetVehicle(queueCmd.LicenseNum);
            int parkingSpaceID = vehiclelist.parkingSpaceID;
            int garageID = vehiclelist.garageID;
            Fetching_Space fslist = oper.GetFetchingSpace(parkingSpaceID);
            int parkingSpaceX = fslist.parkingSpaceX;
            int parkingSpaceY = fslist.parkingSpaceY;
            int parkingSpaceZ = fslist.parkingSpaceZ;
            int status = 4;//取车
            ParkingMessage pm = new ParkingMessage();
            pm.status = status;
            pm.parkingSpaceID = Convert.ToString(parkingSpaceID);
            pm.parkingSpaceX = Convert.ToString(parkingSpaceX);
            pm.parkingSpaceY = Convert.ToString(parkingSpaceY);
            pm.parkingSpaceZ = Convert.ToString(parkingSpaceZ);
            PLC.SetMessage(pm);
            Console.WriteLine("取车:" + queueCmd.LicenseNum);
            Log.WriteLog("取车:" + queueCmd.LicenseNum);
            UILogServer.ins.info("取车:" + queueCmd.LicenseNum);
            //等待PLC取车完成信号
            Task signalFromPLC = Task.Factory.StartNew(() =>
            {
                while (true)
                {
                    Thread.Sleep(2000);
                    PLCMsg = (PLCMessage)PLC.GetMessage();
                    int fetchingStatus = Convert.ToInt32(PLCMsg.plcList[fetch_completed_address].Value);
                    //取车完成信号
                    if (fetchingStatus == 1)
                    {
                        //取车完成后或可归零
                        break;
                    }
                }
            });
            //Task.WaitAll(signalFromPLC);
            signalFromPLC.Wait();
            status = 5;
            ParkingMessage pm2 = new ParkingMessage();
            pm2.status = status;
            PLC.SetMessage(pm2);
            UILogServer.ins.info("号牌："+queueCmd.LicenseNum+"取车完成");
            //数据库更新
            //更新车库表车位数
            int freeSpaceCount = oper.getGarageFreeSpace(garageID);
            freeSpaceCount = freeSpaceCount + 1;
            oper.UpdateGarageFreeSpace(freeSpaceCount, garageID);
            //更新车位表车位状态
            oper.UpdateParkingSpaceState(parkingSpaceID, 0);
            //更新车辆表车辆停车状态
            oper.UpdateVehicleParkState(queueCmd.LicenseNum, 0);
            //更新停车记录表
            oper.UpdateParkingRecords(6, queueCmd.TimeRecord, queueCmd.parkingRecordsID);
        }
    }

    public class ExceptionCmd : AbstractCmd
    {
        public override void executeCmd(Command queueCmd)
        {
            if (queueCmd.commandType == 'e')
            {
                int userId = Convert.ToInt32(queueCmd.userID);
                //过期用户指令
                oper.InsertToMessageQueue(userId, "停车异常，请联系管理员！", 2);
                LogFile logFile = LogFile.ERROR_NUMBERPLATE;
                Log.WriteLog(logFile, "过期用户指令,车牌号：" + queueCmd.LicenseNum);
                UILogServer.ins.error("过期用户指令,车牌号：" + queueCmd.LicenseNum);
                //continue;
            }
        }
    }

    public class SimpleCMDFactory
    {
        public AbstractCmd createCmd(Command queueCmd)
        {
            AbstractCmd abstractCmd = null;

            if (queueCmd.commandType == 's')
            {
                abstractCmd = new StopCmd();
            }
            if (queueCmd.commandType == 'f')
            {
                abstractCmd = new FetchCmd();
            }
            if (queueCmd.commandType == 'e')
            {
                abstractCmd = new ExceptionCmd();
            }
            return abstractCmd;
        }
    }

}
